import { numCpus, makeWorkerCode, spawnWorker, addWorkerListener, terminateWorker, } from './worker';
export class Ework {
    constructor(worker, options = {}) {
        if (typeof worker !== 'function') {
            throw new TypeError('worker must be a function');
        }
        if (typeof options !== 'object' || options === null) {
            throw new TypeError('options must be an object');
        }
        const { maxWorkers = numCpus, minFreeThreads = 1, numWorkers, init, initData, } = options;
        if (numWorkers !== undefined) {
            if (!Number.isInteger(numWorkers) || numWorkers < 1) {
                throw new RangeError('options.numWorkers must be a positive integer');
            }
            this.totalWorkers = numWorkers;
        }
        else {
            if (!Number.isInteger(maxWorkers) || maxWorkers < 1) {
                throw new RangeError('options.maxWorkers must be a positive integer');
            }
            if (!Number.isInteger(minFreeThreads) || minFreeThreads < 0) {
                throw new RangeError('options.minFreeThreads must be a positive integer or 0');
            }
            this.totalWorkers = Math.max(Math.min(maxWorkers, numCpus - minFreeThreads), 1);
        }
        if (typeof init !== 'undefined' && typeof init !== 'function') {
            throw new TypeError('options.init must be a function');
        }
        const initString = init ? init.toString() : '() => null';
        const workerString = worker.toString();
        const workerCode = makeWorkerCode(initString, workerString);
        this.freeWorkers = 0;
        this.workers = [];
        for (let i = 0; i < this.totalWorkers; i++) {
            const workerInstance = spawnWorker(workerCode);
            const workerObj = {
                worker: workerInstance,
                isWorking: true,
                job: null,
            };
            addWorkerListener(workerInstance, 'message', (message) => {
                if (message.type === 'init') {
                    if (message.status === 'success') {
                        workerObj.isWorking = false;
                        this.freeWorkers++;
                        this.run();
                    }
                    else {
                        // TODO handle init error
                    }
                }
                else if (message.type === 'result') {
                    const job = workerObj.job;
                    assert(job !== null);
                    assert(job.id === message.id);
                    if (message.status === 'success') {
                        job.resolve(message.result);
                    }
                    else {
                        job.reject(new Error(message.error));
                    }
                    workerObj.job = null;
                    workerObj.isWorking = false;
                    this.freeWorkers++;
                    this.run();
                }
                else {
                    assert(false);
                }
            });
            workerInstance.postMessage({ type: 'init', value: initData });
            this.workers.push(workerObj);
        }
        this.queue = [];
        this.nextWorkId = 0;
    }
    async execute(value) {
        return this.enqueue(value);
    }
    async terminate() {
        await Promise.all(this.workers.map(terminateWorker));
        this.freeWorkers = 0;
        this.queue = [];
        this.workers = [];
    }
    async map(values) {
        return Promise.all(values.map(async (value) => this.enqueue(value)));
    }
    async enqueue(value) {
        let resolve;
        let reject;
        const promise = new Promise((res, rej) => {
            resolve = res;
            reject = rej;
        });
        const workload = {
            id: this.nextWorkId++,
            value,
            // @ts-ignore
            resolve,
            // @ts-ignore
            reject,
        };
        this.queue.push(workload);
        this.run();
        return promise;
    }
    run() {
        while (this.freeWorkers > 0 && this.queue.length > 0) {
            const nextJob = this.queue.shift();
            assert(nextJob !== undefined);
            const worker = this.getFreeWorker();
            worker.worker.postMessage({
                id: nextJob.id,
                type: 'work',
                value: nextJob.value,
            });
            worker.job = nextJob;
            worker.isWorking = true;
            this.freeWorkers--;
        }
    }
    getFreeWorker() {
        const worker = this.workers.find((w) => !w.isWorking);
        assert(worker !== undefined);
        return worker;
    }
}
function assert(condition) {
    if (!condition)
        throw new Error('UNREACHABLE');
}
//# sourceMappingURL=Ework.js.map